# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.

# Written by Risto Maki-Petays <risto@kaverit.org>.

# BUGS:
#   Endianness checking is broken
#   Not very safe functions
#   Tags are overriden if multiple tags are read

package LW::Object;

use 5.005;

use strict;
use warnings;
use IO::File;
use File::stat;

our $VERSION = "0.1";

=head1 NAME

LW::Object - A Lightwave object loader

=head1 DESCRIPTION

This is an object-oriented library which can load objects made with Newtek(R) Lightwave.

=head1 SYNOPSIS

Use like this:

  use LW::Object;
  my $obj = new LW::Object();
  $obj->open("filename");

=head1 METHODS

=over

=item new

Instantiates an object which holds an object.

=cut

my $fh;
my $points;
my $endian;
my %form;

sub new {
  my $pkg = shift;
  my $self = bless({
#    filename => "",
  }, $pkg);
#  my $self = {};
#  bless $self;
  # a REALLY bad endian check
  if(vec("1234",0,32) == 825373492) {
    $self->{endian} = 1; # N, network, big endian
  } else {
    $self->{endian} = 0; # V, vax, little endian
  }
  return $self;
}

=item open $filename

Opens the specified object

=cut

sub open {
  my $self = shift;
  my $filename = shift;
  my $poly_size = 0, my $buf;
  # initialize a filehandle for the object file
  $self->{fh} = new IO::File;
  $self->{fh}->open($filename) or return 0; # can't open file
  $self->{filesize} = stat($filename)->size or return 0;
  sysread($self->{fh}, $buf, 4);
  if($buf ne 'FORM') { return 0; } # the file is not IFF LWOB
  sysread($self->{fh}, $buf, 4);
  $self->{form}->{size} = LW::u32($buf) or return 0; # FORM size is zero bytes
  # read FORM, FIX check limits of FORM too
  sysread($self->{fh}, $self->{form}->{body}, $self->{form}->{size}) or return 0; # unable to read from file
  # all data is read from the file, let's close the file handle
  $self->{fh}->close; undef($self->{fh});
  # we could close the file here...
  if(substr($self->{form}->{body}, 0, 4) ne 'LWO2') { return 0; } # the file is not IFF LWOB
  my $position = 4;
  my $tagname = "";
  # parse form block, read all tags to correct hashes
  while($position < $self->{form}->{size}) {
    $tagname = substr($self->{form}->{body}, $position, 4);
    $self->{$tagname}->{size} = LW::u32(substr($self->{form}->{body}, $position + 4, 4));
    # print "$tagname: $self->{$tagname}->{size}\n"; THIS IS BROKEN!
    $self->{$tagname}->{body} = substr($self->{form}->{body}, $position + 8, $self->{$tagname}->{size});
    $position += 8 + $self->{$tagname}->{size};
  }
  # form is parsed, let's discard it
  undef($self->{form});
  # fetch point data from object
  for($position = 0; $position < int($self->{PNTS}->{size} / 12); $position++) {
    push @{$self->{points}}, [(LW::f32(substr($self->{PNTS}->{body}, $position*12, 4)), LW::f32(substr($self->{PNTS}->{body}, $position*12+4, 4)), LW::f32(substr($self->{PNTS}->{body}, $position*12+8, 4)))];
  }
  # fetch polydata from object
  for($position = 4; $position < $self->{POLS}->{size}; $position+=2) {
    my @polygon_points = ();
    my $poly_size = LW::u16(substr($self->{POLS}->{body}, $position, 2));
    for(my $counter = 0; $counter < $poly_size; $counter++) {
#      push @polygon_points, LW::u16(substr($self->{POLS}->{body}, $position+=2, 2));
      my $point_index = LW::u16(substr($self->{POLS}->{body}, $position+=2, 2));
      my @pnts = @{$self->{points}[$point_index]};
      push @polygon_points, [@pnts];
    }
    push @{$self->{polygons}->{points}}, [@polygon_points];
  }
  # read the surface id of every polygon
  for($position = 4; $position < $self->{PTAG}->{size}; $position += 4) {
    push @{$self->{polygons}->{surfaces}}, LW::u16(substr($self->{PTAG}->{body}, $position+2, 2));
  }
  # count the normals for each face
  foreach my $polygon_points (@{$self->{polygons}->{points}}) {
    my @normals = (); my @vectors = ();
    # FIX: check if more than 3 vertices
    my @point = (@$polygon_points[0], @$polygon_points[1], @$polygon_points[2]);
    push @vectors, [($point[1]->[0] - $point[2]->[0],
                     $point[1]->[1] - $point[2]->[1],
                     $point[1]->[2] - $point[2]->[2]
                     )],
                   [($point[1]->[0] - $point[0]->[0],
                     $point[1]->[1] - $point[0]->[1],
                     $point[1]->[2] - $point[0]->[2]
                     )];
    @normals = 
                   ($vectors[0][1] * $vectors[1][2] /2 - $vectors[0][2] * $vectors[1][1],
                     $vectors[0][2] * $vectors[1][0] /2 - $vectors[0][0] * $vectors[1][2],
                     $vectors[0][0] * $vectors[1][1] /2 - $vectors[0][1] * $vectors[1][0]
                    );
    my $dist = sqrt(
                    $normals[0] * $normals[0] +
                    $normals[1] * $normals[1] +
                    $normals[2] * $normals[2]
                   );
    $normals[0] /= $dist;
    $normals[1] /= $dist;
    $normals[2] /= $dist;
    push @{$self->{polygons}->{normals}}, [@normals];

  }

  1;
}

sub getgl {
  my $self = shift;
  my ($poly_index, $poly_size, $point_index, $point);
  $poly_index = 0;
  my @objectdata = ();
  if(!@{$self->{polygons}->{points}}) { return ""; }
  my $y = 0;
  foreach my $polygon (@{$self->{polygons}->{points}}) {
    my @polydata = ();
    push @polydata, [
      $self->{polygons}->{normals}[$poly_index][0],
      $self->{polygons}->{normals}[$poly_index][1],
      $self->{polygons}->{normals}[$poly_index][2]
    ];
    foreach(@{$self->{polygons}->{points}[$poly_index]}) {
      push @polydata, [(@$_[0], @$_[1], @$_[2])];
    }
    $poly_index++;
    push @objectdata, [@polydata];
  }
  @objectdata;
}

sub draw {
  my $self = shift;
  my @gl_type = ("", "GL_POINTS", "GL_LINES", "GL_TRIANGLES", "GL_QUADS", "GL_POLYGON");
  my ($poly_index, $poly_size, $point_index, $point, $gl_string);
  $poly_index = 0;
  $gl_string = "";
  if(!@{$self->{polygons}->{points}}) { return ""; }
  foreach my $polygon (@{$self->{polygons}->{points}}) {
    $gl_string .= "# polygon no: $poly_index surface no: $self->{polygons}->{surfaces}[$poly_index]\n";
    $poly_size = @$polygon;
    if($poly_size < 5) {
      $gl_string .= "glBegin($gl_type[$poly_size]);\n";
    } else {
      $gl_string .= "glBegin($gl_type[5]);\n";
    }
    $gl_string .= sprintf("glNormal( %f, %f, %f);\n",
      $self->{polygons}->{normals}[$poly_index][0],
      $self->{polygons}->{normals}[$poly_index][1],
      $self->{polygons}->{normals}[$poly_index][2]
    );
    foreach(@{$self->{polygons}->{points}[$poly_index]}) {
      $gl_string .= "  glVertex( " .
        @$_[0] . ", " .
        @$_[1] . ", " .
        @$_[2] . ", " .
        ");\n";
    }
    $gl_string .= "glEnd();\n\n";
    $poly_index++;
  }
  $gl_string;
}

=item report

Prints a detailed report of the parsed object.

=cut

sub report {
  my $self = shift;
  print "Total number of points: " . @{$self->{points}} . "\n";
  print "Total number of polygons: " . @{$self->{polygons}->{points}} . "\n";
  LW::plineif("  %d of which are lines\n", $self->return_poly_count(1));
  LW::plineif("  %d of which are lines\n", $self->return_poly_count(2));
  LW::plineif("  %d of which are triangles\n", $self->return_poly_count(3));
  LW::plineif("  %d of which are quads\n", $self->return_poly_count(4));
  LW::plineif("  %d of which are complex polygons\n", $self->return_poly_count(-1));
  1;
  print "\n";
}

=item return_poly_count $point_count

Returns the number of polygons consisting of the predetermined number or
points.

B<Note:> on $point_count

    -1 ... return the number of polygons with more than 4 vertices
     0 ... return the total number of polygons
     1 ... return the number of single points
     2 ... return the number of lines
     3 ... return the number of triangles
     4 ... return the number of quads

B<Example:>

    $count = $object->return_poly_count(4);

    $count = $object->return_poly_count(-1);

=cut

# return the polygon count of n verteces
sub return_poly_count {
  my $self = shift;
  my $point_count = shift;
  my $poly_index = 0;
  my $count = 0;
  foreach my $polygon (@{$self->{polygons}->{points}}) {
    if(@$polygon == $point_count and $point_count != -1 or @$polygon > 4 and $point_count == -1) {
      $count++;
    }
  }
  $count;
}

sub point_count : lvalue {
  my $self = shift;
  return $self->{point_count};
}

1;

__END__

